;----------------------------------------------------
; gfx.asm : graphic routines
;
; (c) 2007 Vincent 'MooZ' Cruz
;
; This is a set of function to manipulate
; palettes, sprites, and background.
;
; LICENCE: not my fault if anything burns
;

	.bss

; This actually places a 'TIA' command
; into RAM, from which to execute
; for BLiT to VRAM for SATB transfer
; and other VRAM load functions

_ram_hdwr_tia		.ds	1
_ram_hdwr_tia_src	.ds	2
_ram_hdwr_tia_dest	.ds	2
_ram_hdwr_tia_size	.ds	2
_ram_hdwr_tia_rts	.ds	1

; BAT informations
bat_width  .ds 2
bat_height .ds 1
bat_hmask  .ds 1
bat_vmask  .ds 1

; Map coords
mapbat_bottom   .ds 1
mapbat_top      .ds 1
mapbat_top_base .ds 2

; Screen dimension
scr_width  .ds 1
scr_height .ds 1

; Scroll display list
; There're 4 scroll windows
; As we are doing some kind of double buffering, the scroll window
; datas are doubled.

						; bat coordinate (values are between 0 and 512)
scroll_bat_xl	.ds 8	; -	X (low nibble)
scroll_bat_xh	.ds 8	; -	X (high nibble)
scroll_bat_yl	.ds 8	; -	Y (low nibble)
scroll_bat_yh	.ds 8	; -	Y (high nibble)

scroll_top		.ds 8	; top scanline
scroll_bottom	.ds 8	; bottom scanline
						
scroll_control	.ds 8	; controll (bg/sp enable?)

scroll_list		.ds 4	; list of windows to display
scroll_idx		.ds 1	; number of windows in the list

        .code
;----------------------------------------------------------------------
; name : map_data
;
; description : Map the pointer passed as argument
;               to banks 3-4 ($6000-$9FFF) and save
;               previous bank values to _bx (_bl,_bh).
;
; Warning: Don't forget to call unmapData when you
;          are done or else the old bank values will
;          be lost.
;
; in:  _BL = data bank
;      _SI = data address
;
; out: _BX = old banks
;      _SI = remapped data address
;
map_data:
  ldx    _bl
  ; Save current map banking
  tma    #3
  sta    <_bl
  tma    #4
  sta    <_bh
  ; Map new banks
  txa
  tam   #3
  inc   A
  tam   #4
  ; Remap data address to page 3
  lda   <_si+1
  and   #$1f
  ora   #$60
  sta   <_si+1
  rts

;----------------------------------------------------------------------
; name : unmap_data
;
; description : Restored previously saved bank mapping
;               to banks 3-4 ($6000-$9FFF).
;
; in:  _BX = saved data banks
;
unmap_data:
  lda    <_bl
  tam    #3
  lda    <_bh
  tam    #4
  rts


;----------------------------------------------------------------------
; name :  remap_data
;
; description : Check if the banks needs to be set to
;               the next 8kb
;
; in: _SI = Current to be remapped if nedded
;
remap_data:
  lda    <_bp
  bne    .l1
  ; The pointer crossed the 8kb boundary?
  lda    <_si+1
  bpl    .l1
  ; So _si is using mpr #5, move it back to mpr #4
  sub    #$20
  sta    <_si+1
  ; mpr #3 = mpr #4
  tma    #4
  tam    #3
  ; Set mpr #4 to the next 8kb
  inc    A
  tam    #4
.l1:
  rts

;----------------------------------------------------------------------
; name : set_write
;
; description: set the VDC VRAM write pointer
;
; in :  _DI = VRAM location
;
set_write:
  vreg    #$00
  lda    <_di
  sta    video_data_l
  lda    <_di+1
  sta    video_data_h
  vreg   #$02
  rts

;----------------------------------------------------------------------
; name : load_palette
;
; description : Initialize a palette in VCE with the
;               data passed as argument
;
; in: _AL = index of the first sub-palette (0-31)
;     _BL = data bank
;     _SI = address of data
;     _CL = number of sub-palette to copy
;
load_palette:
  ; Map data
  jsr    map_data

  ; Multiply sub-palette index by 16 to point to the
  ; correct VCE index
  lda    <_al
  stz    <_ah
  asl    A
  asl    A
  asl    A
  asl    A
  rol    <_ah
  ; Set VCE index
  sta    color_reg_l
  lda    <_ah
  sta    color_reg_h

  ; Use TIA, but BLiT 16 words at a time (32 bytes)
  ; Because interrupt must not be deferred too much
  stw    #32, _ram_hdwr_tia_size
  stw    #color_data, _ram_hdwr_tia_dest
.loop_a:
      stw    <_si, _ram_hdwr_tia_src
      jsr    _ram_hdwr_tia
      addw   #32, <_si
      dec    <_cl
      bne    .loop_a

  ; Copy data
  ; Restore bank mapping
  jsr unmap_data
  rts

;----------------------------------------------------------------------
; name : set_bat_size
;
; description : set bg map virtual size
;
; in : A = new size (0-7)
;
set_bat_size:
  and   #$07
  pha
  ; Memory Access Width Register (vreg #9)
  vreg  #9
  pla
  tax
  ; BG map size is store in bits 4 to 6 in vreg #9
  asl   A
  asl   A
  asl   A
  asl   A
  sta   video_data_l
  ; Convert to "real" screen size and mask
  ; -- width : 32,64,128
  lda  .width,X
  sta   bat_width
  stz   bat_width+1
  dec   A
  sta   bat_hmask
  ; -- height : 32,64
  lda  .height,X
  sta   bat_height
  ; -- bat coords
  sta   mapbat_bottom
  stz   mapbat_top
  stz   mapbat_top_base
  stz   mapbat_top_base+1
  dec   A
  sta   bat_vmask

  rts

.width:	 .db $20,$40,$80,$80,$20,$40,$80,$80
.height: .db $20,$20,$20,$20,$40,$40,$40,$40


;----------------------------------------------------------------------
; name : load_bat
;
; description : Transfer BAT data from ROM/RAM to VRAM
;
; in :  _DI = VRAM base address
;       _BL = BAT bank
;       _SI = BAT memory location
;       _CL = nb of column to copy
;       _CH = nb of row
;
loab_bat:
  cly
  ; Map data
  jsr    map_data

.l1:  ; Set VRAM pointer
      jsr    set_write
      ldx    <_cl
.l2:      ; Copy line
          lda    [_si],Y
          sta    video_data_l
          iny

          lda    [_si],Y
          sta    video_data_h
          iny

          ; if y=0, this means that we copied
          ; 256 bytes and can't further use
          ; lda [zp],Y. That's why we make _si
          ; jump to 256 bytes.
          bne    .l3
              inc    <_si+1
.l3:      dex
          bne    .l2

      ; Jump to next line
      jsr    remap_data
      addw   bat_width, <_di
      dec    <_ch
      bne    .l1

  ; Unmap data
  jsr   unmap_data

  rts

;----------------------------------------------------------------------
; name : set_xres
;
; description : Set horizontal screen resolution
;
; in :  _AH = Horizontal screen resoultion
;       _CL = 'blur'
;
        .bss
hsw       .ds 1		; temporary parameters for calculating video registers
hds       .ds 1
hdw       .ds 1
hde       .ds 1
	.code

_vce_tab:	.db	 0, 1, 2
_hsw_tab:	.db	 2, 3, 5
_hds_tab:	.db	18,25,42
_hde_tab:	.db	38,51,82

set_xres:
  ; TODO disable vsync
  lda	<_ah
  sta	<_bh
  lda	<_al
  sta	<_bl		; bx now has x-res

  lsr	<_bh
  ror	<_bl
  lsr	<_bh
  ror	<_bl
  lsr	<_bl		; bl now has x/8

  cly			; offset into numeric tables
  			; 0=low-res, 1=mid-res, 2=high-res

  lda	<_ah
  beq    .xres_calc	; < 256
  cmp	#3
  bhs   .xres_calc

  cmpw	#$10C,<_ax
  blo    .xres_calc	; < 268

  iny
  cmpw	#$164,<_ax
  blo    .xres_calc	; < 356

  iny			; 356 < x < 512

.xres_calc:
  lda	_vce_tab,Y
  ora	<_cl
  sta	color_ctrl	; dot-clock (x-resolution)

  lda	_hsw_tab,Y	; example calc's (using "low-res" numbers)
  sta	hsw		; hsw = $2
  lda	<_bl
  sta	hds		; hds = (x/8) temporarily
  dec	A
  sta	hdw		; hdw = (x/8)-1
  lsr	hds		; hds = (x/16) temporarily

  lda	_hds_tab,Y
  sub	hds
  sta	hds		; hds = 18 - (x/16)

  lda	_hde_tab,Y
  sub	hds
  sub	<_bl		; hde = (38 - ( (18-(x/16)) + (x/8) ))
  sta	hde

.xres_putit:
  vreg	#$0a
  lda	hsw
  sta	video_data_l
  lda	hds
  sta	video_data_h

  vreg	#$0b
  lda	hdw
  sta	video_data_l
  lda	hde
  sta	video_data_h

.xres_err:

  ; TODO enable vsync
  rts

;----------------------------------------------------------------------
; name : set_xres256
;
; description : Set horizontal screen resolution to 256
;
; in :  _CL = 'blur'
;
set_xres256:
    cla
    ora     <_cl
    sta     color_ctrl

    st0     #$0a
    st1     #$02
    st1     #$02

    st0     #$0b
    st1     #$1f
    st1     #$04

    rts

;----------------------------------------------------------------------
; name : set_xres320
;
; description : Set horizontal screen resolution to 320
;
; in :  _CL = 'blur'
;
set_xres320:
    lda     #$01
    ora     <_cl
    sta     color_ctrl

    st0     #$0a
    st1     #$02
    st1     #$04

    st0     #$0b
    st1     #$2a
    st1     #$04

    rts

;----------------------------------------------------------------------
; name : set_xres512
;
; description : Set horizontal screen resolution to 512
;
; in :  _CL = 'blur'
;
set_xres512:
    lda     #$02
    ora     <_cl
    sta     color_ctrl

    st0     #$0a
    st1     #$02
    st1     #$0b

    st0     #$0a
    st1     #$3f
    st1     #$04

    rts

;----------------------------------------------------------------------
; name : scroll
;
; description : Define a scroll window
;
; in :	\1	scroll window (0-3)
;		\2	Bat x coordinate
;		\3	Bat y coordinate
;		\4	Screen window top (raster position)
;		\5	Screen window bottom (raster position)
;		\6	Screen display control
;
scroll .macro
	ldx		\1

	lda		#low(\2)				; bat x (low)
	sta		scroll_bat_xl, X
	lda		#high(\2)				; bat x (high)
	sta		scroll_bat_xh, X

	lda		#low(\3)				; bat y (low)
	sta		scroll_bat_yl, X
	lda		#high(\3)				; bat y (high)
	sta		scroll_bat_yh, X
	
	lda		\4						; top raster position
	sta		scroll_top, X
	
	lda		\5						; bottom raster position
	sta		scroll_bottom, X
	
	lda		\6
	ora		#$01					; enable window
	sta		scroll_control, X
	
	.endm

;----------------------------------------------------------------------
; name : scroll_disable
;
; description : Disable a scrolling window
;
; in :	\1	scroll window (0-3)
;
scroll_disable .macro
	ldx		\1
	lda		scroll_control, X		; Disabe background display
	and		#$fe
	sta		scroll_control, X

	.endm

;----------------------------------------------------------------------
; name : process_scroll
;
; description :
;
; in :
process_scroll:
	clx
	stz		scroll_idx
	
_process_scroll_0:
	lda		scroll_control, X
	and		#$FE						; skip update for this window if
	beq		_process_scroll_3			; background is disabled
	
_process_scroll_1:
	lda		scroll_top, X				; check if top raster position
	cmp		scr_height					; is bigger than screen height
	bcc		_process_scroll_2
	
		lda		scroll_control, X		; disable window
		and		#$FE
		sta		scroll_control, X

		bra		_process_scroll_3
		
_process_scroll_2:

	lda		scroll_bat_xl,   X			; copy data
	sta		scroll_bat_xl+4, X
	lda		scroll_bat_xh,   X
	sta		scroll_bat_xh+4, X
	lda		scroll_bat_yl,   X
	sta		scroll_bat_yl+4, X
	lda		scroll_bat_yh,   X
	sta		scroll_bat_yh+4, X
	lda		scroll_top,   X
	sta		scroll_top+4, X
	lda		scroll_bottom,   X
	sta		scroll_bottom+4, X
	lda		scroll_control,   X
	sta		scroll_control+4, X


	lda		scroll_control,   X
	sta		scroll_control+4, X
	
	ldy		scroll_idx
	txa
	sta		scroll_list, Y				; scroll_list[scroll_idx] = x
	inc		scroll_idx					; ++scroll_idx
	
	phx									; save loop counter
	
	bra		_process_scroll_sort_2

_process_scroll_sort_0:

		ldx		scroll_list-1, Y
		lda		scroll_top, X
		inc		A
		ldx		scroll_list, Y
		cmp		scroll_top, X
		bcc		_process_scroll_sort_1	; swap if the previous element
			lda		scroll_list-1, Y	; is bigger (plus one) than the 
			sta		scroll_list, Y		; current one
			txa
			sta		scroll_list-1, Y
_process_scroll_sort_1:
		dey								; --y
_process_scroll_sort_2:			
		cpy		#1
		bcs		_process_scroll_sort_0	; while(y >= 1)
		
_process_scroll_3:
	plx

	inx
	cpx		#4							; next entry
	bne		_process_scroll_0

	rts
